/**
 * sNX
 */

package library.graphics;

public class Matrix {
    //Pʍs
    public static double[][] identity() {
        double[][] d = {
            {1.0, 0.0, 0.0, 0.0},
            {0.0, 1.0, 0.0, 0.0},
            {0.0, 0.0, 1.0, 0.0},
            {0.0, 0.0, 0.0, 1.0}
        };
        return d;
    }

    //AtBϊ
    public static double[] transform(double[][] m, double[] v) {
        double[] d = new double[4];
        for(int i=0; i<4; i++)
            for(int j=0; j<4; j++)
                d[i] += m[i][j]*v[j];
        return d;
    }
    public static double[] transform3(double[][] m, double[] v) {
        double w = m[3][0]*v[0] + m[3][1]*v[1] + m[3][2]*v[2] + m[3][3];
        double[] d = {
            (m[0][0]*v[0] + m[0][1]*v[1] + m[0][2]*v[2] + m[0][3]) / w,
            (m[1][0]*v[0] + m[1][1]*v[1] + m[1][2]*v[2] + m[1][3]) / w,
            (m[2][0]*v[0] + m[2][1]*v[1] + m[2][2]*v[2] + m[2][3]) / w,
        };
        return d;
    }

    //
    public static void multiply(double[][] m1, double[][] m2) {
        double[][] d = Array.copy(m1);
        for(int i=0; i<4; i++)
            for(int j=0; j<4; j++) {
                m1[i][j] = 0.0;
                for(int k=0; k<4; k++)
                    m1[i][j] += d[i][k]*m2[k][j];
            }
    }

    //]u
    public static void transport(double[][] m) {
        double[][] d = Array.copy(m);
        for(int i=0; i<4; i++)
            for(int j=0; j<4; j++)
                m[i][j] = d[j][i];
    }

    //]
    public static void rotateX(double[][] m, double angle) {
        double sin = Math.sin(angle);
        double cos = Math.cos(angle);
        double[][] d = {
            {1.0, 0.0,  0.0, 0.0},
            {0.0, cos, -sin, 0.0},
            {0.0, sin,  cos, 0.0},
            {0.0, 0.0,  0.0, 1.0}
        };
        multiply(m, d);
    }
    public static void rotateY(double[][] m, double angle) {
        double sin = Math.sin(angle);
        double cos = Math.cos(angle);
        double[][] d = {
            { cos, 0.0, sin, 0.0},
            { 0.0, 1.0, 0.0, 0.0},
            {-sin, 0.0, cos, 0.0},
            { 0.0, 0.0, 0.0, 1.0}
        };
        multiply(m, d);
    }
    public static void rotateZ(double[][] m, double angle) {
        double sin = Math.sin(angle);
        double cos = Math.cos(angle);
        double[][] d = {
            {cos, -sin, 0.0, 0.0},
            {sin,  cos, 0.0, 0.0},
            {0.0,  0.0, 1.0, 0.0},
            {0.0,  0.0, 0.0, 1.0}
        };
        multiply(m, d);
    }

    //gEk
    public static void scale(double[][] m, double sx, double sy, double sz) {
        double[][] d = {
            { sx, 0.0, 0.0, 0.0},
            {0.0,  sy, 0.0, 0.0},
            {0.0, 0.0,  sz, 0.0},
            {0.0, 0.0, 0.0, 1.0}
        };
        multiply(m, d);
    }

    //sړ
    public static void translate(double[][] m, double dx, double dy, double dz) {
        double[][] d = {
            {1.0, 0.0, 0.0,  dx},
            {0.0, 1.0, 0.0,  dy},
            {0.0, 0.0, 1.0,  dz},
            {0.0, 0.0, 0.0, 1.0}
        };
        multiply(m, d);
    }
}