/**
 * fNX
 */

package library.graphics;

public class Model {
    public double[][] vertex, normal_face, normal_vertex;
    public int[][] edge, face;

    public Model() {
    }
    public Model(double[][] vertex, int[][] index) {
        this.vertex = Array.copy(vertex);

        if(index[0].length == 2) {
            edge = Array.copy(index);
        } else {
            int n = 0;
            for(int i=0; i < index.length; i++) {
                if(index[i].length == 3) n++;
                else n+=2;
            }
            face = new int[n][];
            n = 0;
            for(int i=0; i < index.length; i++) {
                if(index[i].length == 3)
                    face[n++] = Array.copy(index[i]);
                else {
                    face[n++] = Array.make(index[i][0], index[i][1], index[i][2]);
                    face[n++] = Array.make(index[i][2], index[i][3], index[i][0]);
                }
            }

            setNormal();
        }
    }
    public Model(double[][] vertex, int[][] edge, int[][] face) {
        this(vertex, face);
        this.edge = Array.copy(edge);
    }

    //
    public void setBox(double width, double height, double depth) {
        double[][] v = {
            { width/2, height/2, depth/2},
            {-width/2, height/2, depth/2},
            {-width/2,-height/2, depth/2},
            { width/2,-height/2, depth/2},
            { width/2, height/2,-depth/2},
            {-width/2, height/2,-depth/2},
            {-width/2,-height/2,-depth/2},
            { width/2,-height/2,-depth/2}
        };
        vertex = v;

        int[][] e = {
            {0, 1}, {1, 2}, {2, 3}, {3, 0},
            {4, 5}, {5, 6}, {6, 7}, {7, 4},
            {0, 4}, {1, 5}, {2, 6}, {3, 7}
        };
        edge = e;

        int[][] f = {
            {0, 1, 2},
            {2, 3, 0},
            {4, 7, 6},
            {6, 5, 4},
            {0, 4, 5},
            {5, 1, 0},
            {1, 5, 6},
            {6, 2, 1},
            {2, 6, 7},
            {7, 3, 2},
            {3, 7, 4},
            {4, 0, 3}
        };
        face = f;

        setNormal();
    }

    //
    public void setCube(double width) {
        setBox(width, width, width);
    }

    //
    public void setPlane(int n, double width, double height) {
        vertex = new double[(n+1)*(n+1)][];
        edge = new int[2*n*(n+1)][];
        face = new int[2*n*n][];

        int k = 0;
        for(int i=0; i<n+1; i++)
            for(int j=0; j<n+1; j++)
                vertex[k++] = Array.make(-width/2 + width/n*j, -height/2 + height/n*i, 0.0);

        k = 0;
        for(int i=0; i<n; i++)
            for(int j=0; j<n; j++) {
                edge[k++] = Array.make(i*(n+1) + j, (i+1)*(n+1) + j);
                edge[k++] = Array.make(i*(n+1) + j, i*(n+1) + j+1);
            }
        for(int i=0; i<n; i++) {
            edge[k++] = Array.make(n*(n+1) + i, n*(n+1) + i+1);
            edge[k++] = Array.make(i*(n+1) + n, (i+1)*(n+1) + n);
        }

        k = 0;
        for(int i=0; i<n; i++)
            for(int j=0; j<n; j++) {
                face[k++] = Array.make(i*(n+1) + j, i*(n+1) + j+1, (i+1)*(n+1) + j+1);
                face[k++] = Array.make((i+1)*(n+1) + j+1, (i+1)*(n+1) + j, i*(n+1) + j);
            }

        setNormal();
    }
    public void setPlane(double width, double height) {
        setPlane(1, width, height);
    }

    //
    public void setSphere(int n, double radius) {
        double[] tr = new double[n+1];
        double[] ty = new double[n+1];
        for(int i=0; i<n+1; i++) {
            tr[i] = radius * Math.sin(Math.PI/n*i);
            ty[i] = radius * Math.cos(Math.PI/n*i);
        }
        tr[0] = 0.0;
        tr[n] = 0.0;

        setRevolution(2*n, tr, ty);
    }

    //g[X
    public void setTorus(int n1, int n2, double radius1, double radius2) {
        double[] tr = new double[n2+1];
        double[] ty = new double[n2+1];
        for(int i=0; i<n2; i++) {
            tr[i] = radius1 + radius2 * Math.sin(2.0*Math.PI/n2*i);
            ty[i] = radius2 * Math.cos(2.0*Math.PI/n2*i);
        }
        tr[n2] = tr[0];
        ty[n2] = ty[0];

        setRevolution(n1, tr, ty);
    }

    //~
    public void setCylinder(int n, double radius, double height) {
        double[] tr = {
            0.0, radius, radius, 0.0
        };
        double[] ty = {
            height/2.0, height/2.0, -height/2.0, -height/2.0
        };

        setRevolution(n, tr, ty);
    }

    //~
    public void setCone(int n, double radius, double height) {
        double[] tr = {
            0.0, radius, 0.0
        };
        double[] ty = {
            height, 0.0, 0.0
        };

        setRevolution(n, tr, ty);
    }

    //]
    public void setRevolution(int n, double[] r, double[] y) {
        int m = r.length;

        if(r[0] == 0.0 && r[m-1] == 0.0) {
            vertex = new double[(m-2) * n + 2][];
            edge = new int[(2*m-3) * n][];
            face = new int[(2*m-4) * n][];

            int cv = 0, ce = 0, cf = 0;

            double cos, sin;
            vertex[cv++] = Array.make(0.0, y[0], 0.0);
            for(int i=0; i<n; i++) {
                sin = Math.sin(2.0*Math.PI/n*i);
                cos = Math.cos(2.0*Math.PI/n*i);
                for(int j=1; j<m-1; j++)
                    vertex[cv++] = Array.make(r[j] * cos, y[j], -r[j] * sin);
            }
            vertex[cv++] = Array.make(0.0, y[m-1], 0.0);

            for(int i=0; i<n-1; i++) {
                edge[ce++] = Array.make(0, i*(m-2) + 1);
                face[cf++] = Array.make(0, i*(m-2) + 1, (i+1)*(m-2) + 1);
                for(int j=1; j<m-2; j++) {
                    edge[ce++] = Array.make(i*(m-2) + j, (i+1)*(m-2) + j);
                    edge[ce++] = Array.make(i*(m-2) + j, i*(m-2) + j+1);
                    face[cf++] = Array.make(i*(m-2) + j, i*(m-2) + j+1, (i+1)*(m-2) + j+1);
                    face[cf++] = Array.make((i+1)*(m-2) + j+1, (i+1)*(m-2) + j, i*(m-2) + j);
                }
                edge[ce++] = Array.make(i*(m-2) + m-2, (i+1)*(m-2) + m-2);
                edge[ce++] = Array.make(i*(m-2) + m-2, (m-2) * n + 1);
                face[cf++] = Array.make(i*(m-2) + m-2, (m-2) * n + 1, (i+1)*(m-2) + m-2);
            }

            edge[ce++] = Array.make(0, (n-1)*(m-2) + 1);
            face[cf++] = Array.make(0, (n-1)*(m-2) + 1, 0*(m-2) + 1);
            for(int j=1; j<m-2; j++) {
                edge[ce++] = Array.make((n-1)*(m-2) + j, 0*(m-2) + j);
                edge[ce++] = Array.make((n-1)*(m-2) + j, (n-1)*(m-2) + j+1);
                face[cf++] = Array.make((n-1)*(m-2) + j, (n-1)*(m-2) + j+1, 0*(m-2) + j+1);
                face[cf++] = Array.make(0*(m-2) + j+1, 0*(m-2) + j, (n-1)*(m-2) + j);
            }
            edge[ce++] = Array.make((n-1)*(m-2) + m-2, 0*(m-2) + m-2);
            edge[ce++] = Array.make((n-1)*(m-2) + m-2, (m-2) * n + 1);
            face[cf++] = Array.make((n-1)*(m-2) + m-2, (m-2) * n + 1, 0*(m-2) + m-2);
        } else if(r[0] == 0.0) {
            vertex = new double[(m-1) * n + 1][];
            edge = new int[(2*m-2) * n][];
            face = new int[(2*m-3) * n][];

            int cv = 0, ce = 0, cf = 0;

            double cos, sin;
            vertex[cv++] = Array.make(0.0, y[0], 0.0);
            for(int i=0; i<n; i++) {
                sin = Math.sin(2.0*Math.PI/n*i);
                cos = Math.cos(2.0*Math.PI/n*i);
                for(int j=1; j<m; j++)
                    vertex[cv++] = Array.make(r[j] * cos, y[j], -r[j] * sin);
            }

            for(int i=0; i<n-1; i++) {
                edge[ce++] = Array.make(0, i*(m-1) + 1);
                face[cf++] = Array.make(0, i*(m-1) + 1, (i+1)*(m-1) + 1);
                for(int j=1; j<m-1; j++) {
                    edge[ce++] = Array.make(i*(m-1) + j, (i+1)*(m-1) + j);
                    edge[ce++] = Array.make(i*(m-1) + j, i*(m-1) + j+1);
                    face[cf++] = Array.make(i*(m-1) + j, i*(m-1) + j+1, (i+1)*(m-1) + j+1);
                    face[cf++] = Array.make((i+1)*(m-1) + j+1, (i+1)*(m-1) + j, i*(m-1) + j);
                }
                edge[ce++] = Array.make(i*(m-1) + m-1, (i+1)*(m-1) + m-1);
            }

            edge[ce++] = Array.make(0, (n-1)*(m-1) + 1);
            face[cf++] = Array.make(0, (n-1)*(m-1) + 1, 0*(m-1) + 1);
            for(int j=1; j<m-1; j++) {
                edge[ce++] = Array.make((n-1)*(m-1) + j, 0*(m-1) + j);
                edge[ce++] = Array.make((n-1)*(m-1) + j, (n-1)*(m-1) + j+1);
                face[cf++] = Array.make((n-1)*(m-1) + j, (n-1)*(m-1) + j+1, 0*(m-1) + j+1);
                face[cf++] = Array.make(0*(m-1) + j+1, 0*(m-1) + j, (n-1)*(m-1) + j);
            }
            edge[ce++] = Array.make((n-1)*(m-1) + m-1, 0*(m-1) + m-1);
        } else if(r[m-1] == 0.0) {
            double[] tr = new double[m];
            double[] ty = new double[m];

            for(int i=0; i<m; i++) {
                tr[(m-1) - i] = r[i];
                ty[(m-1) - i] = y[i];
            }

            setRevolution(n, tr, ty);
        } else if(r[0] == r[m-1]) {
            vertex = new double[(m-1) * n][];
            edge = new int[(2*m-2) * n][];
            face = new int[(2*m-2) * n][];

            int cv = 0, ce = 0, cf = 0;

            double cos, sin;
            for(int i=0; i<n; i++) {
                sin = Math.sin(2.0*Math.PI/n*i);
                cos = Math.cos(2.0*Math.PI/n*i);
                for(int j=0; j<m-1; j++)
                    vertex[cv++] = Array.make(r[j] * cos, y[j], -r[j] * sin);
            }

            for(int i=0; i<n-1; i++) {
                for(int j=0; j<m-2; j++) {
                    edge[ce++] = Array.make(i*(m-1) + j, (i+1)*(m-1) + j);
                    edge[ce++] = Array.make(i*(m-1) + j, i*(m-1) + j+1);
                    face[cf++] = Array.make(i*(m-1) + j, i*(m-1) + j+1, (i+1)*(m-1) + j+1);
                    face[cf++] = Array.make((i+1)*(m-1) + j+1, (i+1)*(m-1) + j, i*(m-1) + j);
                }
                edge[ce++] = Array.make(i*(m-1) + m-2, (i+1)*(m-1) + m-2);
                edge[ce++] = Array.make(i*(m-1) + m-2, i*(m-1) + 0);
                face[cf++] = Array.make(i*(m-1) + m-2, i*(m-1) + 0, (i+1)*(m-1) + 0);
                face[cf++] = Array.make((i+1)*(m-1) + 0, (i+1)*(m-1) + m-2, i*(m-1) + m-2);
            }

            for(int j=0; j<m-2; j++) {
                edge[ce++] = Array.make((n-1)*(m-1) + j, 0*(m-1) + j);
                edge[ce++] = Array.make((n-1)*(m-1) + j, (n-1)*(m-1) + j+1);
                face[cf++] = Array.make((n-1)*(m-1) + j, (n-1)*(m-1) + j+1, 0*(m-1) + j+1);
                face[cf++] = Array.make(0*(m-1) + j+1, 0*(m-1) + j, (n-1)*(m-1) + j);
            }
            edge[ce++] = Array.make((n-1)*(m-1) + m-2, 0*(m-1) + m-2);
            edge[ce++] = Array.make((n-1)*(m-1) + m-2, (n-1)*(m-1) + 0);
            face[cf++] = Array.make((n-1)*(m-1) + m-2, (n-1)*(m-1) + 0, 0*(m-1) + 0);
            face[cf++] = Array.make(0*(m-1) + 0, 0*(m-1) + m-2, (n-1)*(m-1) + m-2);

            System.out.println("ok");
        } else {
            vertex = new double[m * n][];
            edge = new int[(2*m-1) * n][];
            face = new int[(2*m-2) * n][];

            int cv = 0, ce = 0, cf = 0;

            double cos, sin;
            for(int i=0; i<n; i++) {
                sin = Math.sin(2.0*Math.PI/n*i);
                cos = Math.cos(2.0*Math.PI/n*i);
                for(int j=0; j<m; j++)
                    vertex[cv++] = Array.make(r[j] * cos, y[j], -r[j] * sin);
            }

            for(int i=0; i<n-1; i++) {
                for(int j=0; j<m-1; j++) {
                    edge[ce++] = Array.make(i*m + j, (i+1)*m + j);
                    edge[ce++] = Array.make(i*m + j, i*m + j+1);
                    face[cf++] = Array.make(i*m + j, i*m + j+1, (i+1)*m + j+1);
                    face[cf++] = Array.make((i+1)*m + j+1, (i+1)*m + j, i*m + j);
                }
                edge[ce++] = Array.make(i*m + m-1, (i+1)*m + m-1);
            }

            for(int j=0; j<m-1; j++) {
                edge[ce++] = Array.make((n-1)*m + j, 0*m + j);
                edge[ce++] = Array.make((n-1)*m + j, (n-1)*m + j+1);
                face[cf++] = Array.make((n-1)*m + j, (n-1)*m + j+1, 0*m + j+1);
                face[cf++] = Array.make(0*m + j+1, 0*m + j, (n-1)*m + j);
            }
            edge[ce++] = Array.make((n-1)*m + m-1, 0*m + m-1);
        }

        setNormal();
    }

    public void setNormal() {
        normal_face = new double[face.length][];
        normal_vertex = new double[vertex.length][3];

        for(int i=0; i < face.length; i++) {
            normal_face[i] = Vector.cross(Vector.sub(vertex[face[i][1]], vertex[face[i][0]]),
                    Vector.sub(vertex[face[i][2]], vertex[face[i][0]]));
            for(int j=0; j<3; j++)
                normal_vertex[face[i][j]] = Vector.add(normal_vertex[face[i][j]], normal_face[i]);
        }

        double[] cvertex = new double[3];
        for(int i=0; i < face.length; i++) {
            for(int j=0; j<3; j++)
                cvertex[j] = (vertex[face[i][0]][j] + vertex[face[i][1]][j] + vertex[face[i][2]][j]) / 3.0;
            normal_face[i] = Vector.add(Vector.normalize(normal_face[i]), cvertex);
        }
        for(int i=0; i < vertex.length; i++)
            normal_vertex[i] = Vector.add(Vector.normalize(normal_vertex[i]), vertex[i]);
    }
}