/**
 * _[i{[jNX
 * Created on 2004/09/26
 * @author Takuya (http://clustera.skr.jp/java/)
 * @version 1.5
 */

package library.graphics;

import java.awt.*;
import java.awt.image.*;

public class VRenderer {
    private int[] pbuf, pbuf2;           //sNZobt@
    private int width, resolution;       //ʂ̕A𑜓x
    private double step, distance;       //TvOԊuA_܂ł̋
    private double intensity, reduction; //{NZ̖x̘ǎWɕϊWǍ̋x
    private boolean cube_visible;        //g̗L

    private double[] eye, scr, scr_dx, scr_dy;
    private double[][] matEye, matCube;
    private Voxel voxel;
    private Component c;
    private MemoryImageSource mis;
    private Image image;
    private Model cube;

    public VRenderer(Component c, int width) {
        this.c = c;
        this.width = width;

        pbuf = new int[width*width];
        pbuf2 = new int[width*width];
        for(int i=0; i < width*width; i++)
            pbuf[i] = (255 << 24);

        mis = new MemoryImageSource(width, width, pbuf, 0, width);
        mis.setAnimated(true);
        image = c.createImage(mis);

        cube = new Model();
        cube.setCube(1.0);
    }

    public void draw(Graphics g) {
        mis.newPixels();
        g.drawImage(image, 0, 0, c);
    }

    public void setLookAt(double distance, double theta, double phi) {
        this.distance = distance;

        matEye = Matrix.identity();
        Matrix.rotateZ(matEye, phi);
        Matrix.rotateY(matEye, theta);

        matCube = Matrix.identity();
        Matrix.rotateY(matCube, -theta);
        Matrix.rotateZ(matCube, -phi);

        eye = Matrix.transform3(matEye, Array.make(0.0, 0.0, distance));
        scr = Matrix.transform3(matEye, Array.make(-0.5, -0.5, 1.0));
        scr_dx = Vector.sub(Matrix.transform3(matEye, Array.make(-0.5, 0.5, 1.0)), scr);
        scr_dy = Vector.sub(Matrix.transform3(matEye, Array.make(0.5, -0.5, 1.0)), scr);
    }

    public void setLight(double intensity, double reduction) {
        this.intensity = intensity;
        this.reduction = reduction;
    }

    public void setQuality(int resolution, double step) {
        this.resolution = resolution;
        this.step = step;
    }

    public void setCube(boolean visible) {
        cube_visible = visible;
    }

    public void raytrace(Voxel voxel) {
        this.voxel = voxel;
            //long t = System.currentTimeMillis();
        raytrace();
            //System.out.println(System.currentTimeMillis() - t);
        if(cube_visible) drawCube();
    }

    //Cg[VO
    private void raytrace() {
        int ratio = width/resolution;
        for(int i=0; i < resolution; i++)
            for(int j=0; j < resolution; j++)
                pbuf2[i + j*resolution] = raycast((double)i/resolution, (double)j/resolution);
        for(int i=0; i < width; i++)
            for(int j=0; j < width; j++)
                pbuf[i + j*width] = pbuf2[(i/ratio) + (j/ratio)*resolution];
    }

    private int raycast(double dx, double dy) {
        double sum_density = 0.0, sum_red = 0.0, sum_blue = 0.0, density;
        double[] ray = Vector.add(scr, Vector.add(Vector.scale(scr_dx, dx), Vector.scale(scr_dy, dy)));
        double[] dv = Vector.scale(Vector.normalize(Vector.sub(ray, eye)), step);

        for(int i = (int)(2.0/step); !voxel.inner(ray); i--) {
            ray = Vector.add(ray, dv);
            if(i == 0) return (255<<24);
        }

        while(voxel.inner(ray)) {
            density = voxel.getDensity(ray);
            ray = Vector.add(ray, dv);
            if(density > 0) {
                sum_density += density;
                sum_red += intensity * density * Math.exp(- reduction * sum_density*step);
            } else {
                sum_density -= density;
                sum_blue -= intensity * density * Math.exp(- reduction * sum_density*step);
            }
        }

        int red = (int)(sum_red*step * 256);
        int blue = (int)(sum_blue*step * 256);
        if(red > 255) red = 255;
        if(blue > 255) blue = 255;

        return (255<<24) | (red<<16) | blue;
    }

    //̘̂g\
    private void drawCube() {
        for(int i=0; i < cube.edge.length; i++) {
            int[][] iv = new int[2][3];
            for(int j=0; j<2; j++) {
                double[] v = Matrix.transform3(matCube, cube.vertex[cube.edge[i][j]]);
                iv[j][0] = (int)(v[1] * (distance - 1.0)/(distance  - v[2]) * width) + width/2;
                iv[j][1] = (int)(v[0] * (distance - 1.0)/(distance  - v[2]) * width) + width/2;
                iv[j][2] = (int)((v[2] + 1.0) * 0x40000000);
            }
            drawLine(iv);
        }
    }

    private void drawLine(int[][] tv) {
        int x, y, z, dx, dy, dz, dx2, dy2, ex, ey, wx, wy;

        x = tv[0][0];
        y = tv[0][1];
        z = tv[0][2];
        dx = tv[1][0] - tv[0][0];
        dy = tv[1][1] - tv[0][1];
        dz = tv[1][2] - tv[0][2];
        dx2 = 2*Math.abs(dx);
        dy2 = 2*Math.abs(dy);
        wx = Math.abs(dx) + 1;
        wy = Math.abs(dy) + 1;

        if(wx > wy) {
            ey = -Math.abs(dx);
            dz /= wx;
            for(int i=0; i<wx; i++) {
                plot(x, y, (z>>23));
                x += addDir(dx);
                ey += dy2;
                z += dz;
                if(ey >= 0) {
                    y += addDir(dy);
                    ey -= dx2;
                }
            }
        } else {
            ex = -Math.abs(dy);
            dz /= wy;
            for(int i=0; i<wy; i++) {
                plot(x, y, (z>>23));
                y += addDir(dy);
                ex += dx2;
                z += dz;
                if(ex >= 0) {
                    x += addDir(dx);
                    ex -= dy2;
                }
            }
        }
    }
    private int addDir(int dxy) {
        if(dxy < 0) return -1;
        else if(dxy == 0) return 0;
        else return 1;
    }

    private void plot(int x, int y, int z) {
        if(x >= 0 && x < width && y >= 0 && y < width) {
            int c = pbuf[x + y*width];
            int r = z + ((c & 0x00FF0000) >> 16);
            int g = z + ((c & 0x0000FF00) >> 8);
            int b = z +  (c & 0x000000FF);
            if(r > 255) r = 255;
            if(g > 255) g = 255;
            if(b > 255) b = 255;

            pbuf[x + y*width] = (255<<24) | (r<<16) | (g<<8) | b;
        }
    }
}